#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from __future__ import annotations

from gettext import gettext as tr
from typing import TYPE_CHECKING

from holidays.calendars import _CustomIslamicHolidays
from holidays.calendars.gregorian import MAR, JUN, JUL, AUG
from holidays.groups import ChristianHolidays, InternationalHolidays, IslamicHolidays
from holidays.observed_holiday_base import (
    ObservedHolidayBase,
    SAT_TO_NEXT_WORKDAY,
    SAT_SUN_TO_NEXT_WORKDAY,
)

if TYPE_CHECKING:
    from datetime import date


class Rwanda(ObservedHolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """Rwanda holidays.

    References:
        * <https://en.wikipedia.org/wiki/Public_holidays_in_Rwanda>
        * [Presidential Order N° 06/01 of 16/02/2011](https://web.archive.org/web/20250601161219/https://archive.gazettes.africa/archive/rw/2011/rw-government-gazette-dated-2011-03-07-no-10.pdf)
        * [Presidential Order N° 42/03 of 30/06/2015](https://web.archive.org/web/20180417062635/http://www.igihe.com/IMG/pdf/iteka_rya_perezida_rigena_iminsi_y_ikiruhuko_rusange.pdf)
        * [Presidential Order N° 54/01 of 24/02/2017](https://web.archive.org/web/20220626143357/https://www.ngoma.gov.rw/index.php?eID=dumpFile&t=f&f=44336&token=fc82c76109af7950f8895d40ddd3e15bd8c57e8c)
        * [Presidential Order N° 62/01 of 19/10/2022](https://web.archive.org/web/20250815032533/https://mifotra.prod.risa.rw/index.php?eID=dumpFile&t=f&f=85230&token=bcf0bf166638c11f4bf3d2f8c629052db26d38b6)
    """

    country = "RW"
    default_language = "rw"
    # %s (estimated).
    estimated_label = tr("%s (yagereranijwe)")
    # %s (observed).
    observed_label = tr("%s (yizihijwe)")
    # %s (observed, estimated).
    observed_estimated_label = tr("%s (yizihijwe, yagereranijwe)")
    supported_languages = ("en_US", "fr", "rw")
    # Presidential Order N° 06/01 of 16/02/2011.
    start_year = 2012

    def __init__(self, *args, islamic_show_estimated: bool = True, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(
            self, cls=RwandaIslamicHolidays, show_estimated=islamic_show_estimated
        )
        kwargs.setdefault("observed_since", 2017)
        super().__init__(*args, **kwargs)

    def _populate_observed(self, dts: set[date], *, multiple: bool = False) -> None:
        """
        Applies `SAT_TO_NEXT_WORKDAY` instead of `SAT_SUN_TO_NEXT_WORKDAY`
        observed_rule for Day after New Year's Day and Boxing Day.
        """
        special_cases = {
            # Day after New Year's Day.
            self.tr("Umunsi ukurikira Ubunani"),
            # Boxing Day.
            self.tr("Umunsi ukurikira Noheli"),
        }
        for dt in sorted(dts):
            for name in self.get_list(dt):
                self._add_observed(
                    dt,
                    name,
                    rule=SAT_TO_NEXT_WORKDAY if name in special_cases else SAT_SUN_TO_NEXT_WORKDAY,
                )

    def _populate_public_holidays(self):
        dts_observed = set()

        # New Year's Day.
        dt = self._add_new_years_day(tr("Ubunani"))
        if self._year >= 2018:
            dts_observed.add(dt)

        # Added via Presidential Order N° 42/03 of 30/06/2015.
        if self._year >= 2016:
            # Day after New Year's Day.
            dts_observed.add(self._add_new_years_day_two(tr("Umunsi ukurikira Ubunani")))

        # National Heroes' Day.
        dts_observed.add(self._add_holiday_feb_1(tr("Umunsi w'Intwari")))

        # Memorial Day of Genocide perpetrated against the Tutsi in 1994.
        self._add_holiday_apr_7(tr("Umunsi wo Kwibuka Jenoside yakorewe Abatutsi mu 1994"))

        # Good Friday.
        self._add_good_friday(tr("Umunsi wa Gatanu Mutagatifu"))

        # Added via Presidential Order N° 54/01 of 24/02/2017.
        if self._year >= 2017:
            # Easter Monday.
            self._add_easter_monday(tr("Ku wa mbere wa Pasika"))

        # Labor Day.
        dts_observed.add(self._add_labor_day(tr("Umunsi Mukuru w'Umurimo")))

        # Independence Day.
        dts_observed.add(self._add_holiday_jul_1(tr("Umunsi w'Ubwigenge")))

        # Liberation Day.
        dts_observed.add(self._add_holiday_jul_4(tr("Umunsi wo Kwibohora")))

        # Added via Presidential Order N° 42/03 of 30/06/2015.
        if self._year >= 2015:
            # Umuganura Day.
            self._add_holiday_1st_fri_of_aug(tr("Umunsi w'Umuganura"))

        # Assumption Day.
        dts_observed.add(self._add_assumption_of_mary_day(tr("Ijyanwa mu Ijuru rya Bikiramariya")))

        # Christmas Day.
        dts_observed.add(self._add_christmas_day(tr("Noheli")))

        # Boxing Day.
        dts_observed.add(self._add_christmas_day_two(tr("Umunsi ukurikira Noheli")))

        # Eid al-Fitr.
        dts_observed.update(self._add_eid_al_fitr_day(tr("Eid El Fitr")))

        # Added via Presidential Order N° 42/03 of 30/06/2015.
        if self._year >= 2015:
            # Eid al-Adha.
            dts_observed.update(self._add_eid_al_adha_day(tr("Eid al-Adha")))

        if self.observed:
            self._populate_observed(dts_observed)


class RW(Rwanda):
    pass


class RWA(Rwanda):
    pass


class RwandaIslamicHolidays(_CustomIslamicHolidays):
    EID_AL_ADHA_DATES_CONFIRMED_YEARS = (2018, 2025)
    EID_AL_ADHA_DATES = {
        2018: (AUG, 22),
        2019: (AUG, 12),
        2022: (JUL, 11),
    }

    EID_AL_FITR_DATES_CONFIRMED_YEARS = (2012, 2025)
    EID_AL_FITR_DATES = {
        2014: (JUL, 29),
        2015: (JUL, 18),
        2016: (JUL, 7),
        2017: (JUN, 26),
        2025: (MAR, 31),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars import _CustomIslamicHolidays
from holidays.calendars.ethiopian import ETHIOPIAN_CALENDAR, is_ethiopian_leap_year
from holidays.calendars.gregorian import AUG, SEP, NOV
from holidays.constants import PUBLIC, WORKDAY
from holidays.groups import ChristianHolidays, InternationalHolidays, IslamicHolidays
from holidays.holiday_base import HolidayBase


class Ethiopia(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """Ethiopia holidays.

    References:
        * [Proclamation No. 29/1996](https://web.archive.org/web/20240918183943/https://lawethiopia.com/images/federal_proclamation/proclamations_by_number/29.pdf)
        * [Proclamation No. 1334/2024](https://web.archive.org/web/20240823080839/https://lawethiopiacomment.wordpress.com/wp-content/uploads/2024/08/public-holiday-stamped.pdf)
        * <https://web.archive.org/web/20250427173714/https://www.edarabia.com/ethiopia/public-holidays/>
        * <https://en.wikipedia.org/wiki/Public_holidays_in_Ethiopia>
        * <https://en.wikipedia.org/wiki/Nations,_Nationalities_and_Peoples'_Day>
        * <https://web.archive.org/web/20250811041201/https://www.timeanddate.com/holidays/ethiopia/2025>
    """

    country = "ET"
    default_language = "am"
    # %s (estimated).
    estimated_label = tr("%s (ግምት)")
    # Negarit Gazeta Proclamation No. 16/1975.
    start_year = 1976
    supported_categories = (PUBLIC, WORKDAY)
    supported_languages = ("am", "ar", "en_ET", "en_US")

    def __init__(self, *args, islamic_show_estimated: bool = True, **kwargs):
        """
        Args:
            islamic_show_estimated:
                Whether to add "estimated" label to Islamic holidays name
                if holiday date is estimated.
        """
        ChristianHolidays.__init__(self, ETHIOPIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(
            self, cls=EthiopiaIslamicHolidays, show_estimated=islamic_show_estimated
        )
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # Christmas Day.
        self._add_christmas_day(tr("የገና ወይም የልደት በዓል"))

        # Epiphany.
        self._add_epiphany_day(tr("የጥምቀት በዓል"))

        if self._year >= 1996:
            # Adwa Victory Day.
            self._add_holiday_mar_2(tr("የአድዋ ድል በዓል"))

        # Good Friday.
        self._add_good_friday(tr("የስቅለት በዓል"))

        # Easter Sunday.
        self._add_easter_sunday(tr("የትንሳኤ(ፋሲካ) በዓል"))

        # International Workers' Day.
        self._add_labor_day(tr("የዓለም የሠራተኞች (የላብአደሮች) ቀን"))

        # Ethiopian Patriots' Victory Day.
        self._add_holiday_may_5(tr("የአርበኞች (የድል) ቀን በዓል"))

        if self._year >= 1992:
            # Downfall of the Dergue Regime Day.
            self._add_holiday_may_28(tr("ደርግ የወደቀበት ቀን"))

        # Ethiopian New Year.
        self._add_ethiopian_new_year(tr("የዘመን መለወጫ (እንቁጣጣሽ) በዓል"))

        # Finding of True Cross.
        self._add_finding_of_true_cross(tr("የመስቀል በዓል"))

        if self._year <= 1990:
            # Julian Date Drift shouldn't affect this one.
            self._add_holiday(
                # Popular Revolution Commemoration Day.
                tr("የአብዮት ቀን"),
                SEP,
                13 if is_ethiopian_leap_year(self._year) else 12,
            )

            # October Revolution Day.
            self._add_holiday_nov_7(tr("የጥቅምት አብዮት ቀን"))

        # Eid al-Fitr.
        self._add_eid_al_fitr_day(tr("የኢድ አልፈጥር"))

        # Eid al-Adha.
        self._add_eid_al_adha_day(tr("የኢድ አልአድሃ (አረፋ)"))

        # Prophet's Birthday.
        self._add_mawlid_day(tr("የመውሊድ በዓል"))

    def _populate_workday_holidays(self):
        # Ethiopian Martyrs' Day.
        self._add_holiday_feb_20(tr("የሰማዕታት ቀን"))

        if self._year >= 2006:
            # Nations, Nationalities and Peoples Day.
            self._add_holiday_dec_9(tr("የብሔር ብሔረሰቦች ቀን"))


class ET(Ethiopia):
    pass


class ETH(Ethiopia):
    pass


class EthiopiaIslamicHolidays(_CustomIslamicHolidays):
    EID_AL_ADHA_DATES_CONFIRMED_YEARS = (2018, 2025)
    EID_AL_ADHA_DATES = {
        2018: (AUG, 22),
    }

    EID_AL_FITR_DATES_CONFIRMED_YEARS = (2018, 2025)

    MAWLID_DATES_CONFIRMED_YEARS = (2018, 2025)
    MAWLID_DATES = {
        2018: (NOV, 21),
        2019: (NOV, 10),
        2025: (SEP, 5),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from datetime import date

GREGORIAN_CALENDAR = "GREGORIAN_CALENDAR"

MON, TUE, WED, THU, FRI, SAT, SUN = range(7)
WEEKEND = (SAT, SUN)

JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC = range(1, 13)

DAYS = {str(d) for d in range(1, 32)}
MONTHS = {
    m: i
    for i, m in enumerate(
        ("jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"), 1
    )
}
WEEKDAYS = {w: i for i, w in enumerate(("mon", "tue", "wed", "thu", "fri", "sat", "sun"))}


# Holiday names.
CHRISTMAS = "christmas"
WINTER_SOLSTICE = "winter_solstice"


def _timedelta(dt: date, days: int = 0) -> date:
    """
    Return date that is `days` days after (days > 0) or before (days < 0) specified date.
    """

    return date.fromordinal(dt.toordinal() + days)


def _get_nth_weekday_from(n: int, weekday: int, from_dt: date) -> date:
    """
    Return date of a n-th weekday before a specific date
    if n is negative.
    Return date of n-th weekday after (including) a specific date
    if n is positive.
    Examples: 1st Monday, 2nd Saturday, etc).
    """

    return _timedelta(
        from_dt,
        (
            (n - 1) * 7 + (weekday - from_dt.weekday()) % 7
            if n > 0
            else (n + 1) * 7 - (from_dt.weekday() - weekday) % 7
        ),
    )


def _get_nth_week_of_month(n: int, month: int, year: int) -> date:
    """
    Return the Monday at the start of week ``n`` for ``month`` in ``year``.

    Weeks span Monday through Sunday. If ``n`` is greater than ``0``, then week ``1``
    contains day ``1`` of ``month``. That Monday may fall in the preceding calendar
    month; each higher ``n`` advances one week chronologically.

    If ``n`` is negative, numbering runs backward from the end of the month: ``-1`` means
    the Monday whose Sunday is the latest Sunday that still falls in ``month`` (the Monday
    may itself fall in the preceding month); ``-2`` is one week earlier, etc.

    Raises ``ValueError`` when no such ``n`` appears in ``month``.
    """
    original_n, original_month, original_year = n, month, year
    if n < 0:
        month += 1
        if month > 12:
            month = 1
            year += 1
        n += 1

    start_date = date(year, month, 1)
    dt = _timedelta(start_date, -start_date.weekday() + (n - 1) * 7)

    # Reject out-of-range n when the week's Sunday is not in the requested year+month.
    week_end = _timedelta(dt, +6)
    if (week_end.year, week_end.month) != (original_year, original_month):
        raise ValueError(
            f"Month {original_month} of {original_year} does not have week {original_n}"
        )

    return dt


def _get_nth_weekday_of_month(n: int, weekday: int, month: int, year: int) -> date:
    """
    Return date of n-th weekday of month for a specific year
    (e.g. 1st Monday of Apr, 2nd Friday of June, etc).
    If n is negative the countdown starts at the end of month
    (i.e. -1 is last).
    """

    requested_year_month = (year, month)

    if n < 0:
        month += 1
        if month > 12:
            month = 1
            year += 1
        start_date = _timedelta(date(year, month, 1), -1)
    else:
        start_date = date(year, month, 1)

    dt = _get_nth_weekday_from(n, weekday, start_date)
    dt_year_month = (dt.year, dt.month)

    if dt_year_month != requested_year_month:
        raise ValueError(f"{dt_year_month} returned for {requested_year_month}")

    return dt


def _get_all_sundays(year):
    first_sunday = _get_nth_weekday_of_month(1, SUN, JAN, year)
    for n in range(0, (date(year, DEC, 31) - first_sunday).days + 1, 7):
        yield _timedelta(first_sunday, n)



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays import HolidayBase
from holidays.calendars import _CustomHinduHolidays, _CustomIslamicHolidays
from holidays.calendars.gregorian import MAR, APR, JUN, JUL, AUG, SEP, OCT, NOV
from holidays.groups import (
    ChineseCalendarHolidays,
    ChristianHolidays,
    HinduCalendarHolidays,
    InternationalHolidays,
    IslamicHolidays,
)


class Suriname(
    HolidayBase,
    ChineseCalendarHolidays,
    ChristianHolidays,
    HinduCalendarHolidays,
    InternationalHolidays,
    IslamicHolidays,
):
    """Suriname holidays.

    References:
        * [Besluit Vrije Dagen 1971](https://web.archive.org/web/20250411201210/https://www.sris.sr/wp-content/uploads/2022/10/Besluit-Vrije-dagen-GB-1971-no-78.pdf)
        * https://web.archive.org/web/20240416223555/https://gov.sr/wp-content/uploads/2022/09/arbeidswet.pdf
        * [S.B. 2007 no. 98](https://web.archive.org/web/20250411204836/https://www.sris.sr/wp-content/uploads/2021/10/SB-2007-no-98-besluit-vrije-dagen.pdf)
        * [S.B. 2012 no. 21](https://web.archive.org/web/20250411204754/https://www.sris.sr/wp-content/uploads/2021/10/SB-2012-no-21-besluit-vrije-dagen.pdf)
        * [S.B. 2021 no. 27](https://web.archive.org/web/20250411210516/https://www.sris.sr/wp-content/uploads/2021/03/S.B.-2021-no.-27-houdende-nadere-wijziging-van-het-Besluit-Vrije-Dagen-1971.pdf)
        * [S.B. 2024 no. 167](https://web.archive.org/web/20250411211129/https://www.sris.sr/wp-content/uploads/2025/01/S.B.-2024-no.-167-Algemene-termijnenwet.pdf)

    Note:
        The oldest decree available online that underpins the public holidays defined here
        for Suriname is Besluit Vrije Dagen 1971 of April 22, 1971.

        The S.B. 2024 no. 167 law only applies to prolongations of terms specified in contracts
        and other similar legal documents, and does not apply to labor agreements (and days off).
    """

    country = "SR"
    default_language = "nl"
    # %s (estimated).
    estimated_label = tr("%s (geschat)")
    start_year = 1972
    supported_languages = ("en_US", "nl")

    def __init__(self, *args, islamic_show_estimated: bool = True, **kwargs):
        """
        Args:
            islamic_show_estimated:
                Whether to add "estimated" label to Islamic holidays name
                if holiday date is estimated.
        """
        ChineseCalendarHolidays.__init__(self)
        ChristianHolidays.__init__(self)
        HinduCalendarHolidays.__init__(self, cls=SurinameHinduHolidays)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(
            self, cls=SurinameIslamicHolidays, show_estimated=islamic_show_estimated
        )
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # New Year's Day.
        self._add_new_years_day(tr("Nieuwjaarsdag"))

        if 1981 <= self._year <= 1992 or 2012 <= self._year <= 2020:
            # Day of Liberation and Renewal.
            self._add_holiday_feb_25(tr("Dag van Bevrijding en Vernieuwing"))

        # Good Friday.
        self._add_good_friday(tr("Goede Vrijdag"))

        # Easter Monday.
        self._add_easter_monday(tr("Tweede Paasdag"))

        if self._year <= 1975:
            # Birthday of H.M. the Queen.
            self._add_holiday_apr_30(tr("Verjaardag van H.M. de Koningin"))

        # Labor Day.
        self._add_labor_day(tr("Dag van de Arbeid"))

        self._add_holiday_jul_1(
            # Day of Freedoms.
            tr("Keti Koti Dey") if 2008 <= self._year <= 2024 else tr("Dag der Vrijheden")
        )

        if self._year >= 2007:
            # Indigenous People Day.
            self._add_holiday_aug_9(tr("Dag der Inheemsen"))

        if self._year >= 2012:
            # Day of the Maroons.
            self._add_holiday_oct_10(tr("Dag der Marrons"))

        if self._year >= 1976:
            self._add_holiday_nov_25(
                # Independence Day.
                tr("Onafhankelijkheidsdag")
                if self._year >= 2008
                # Republic Day.
                else tr("Dag van de Republiek")
            )

        # Christmas Day.
        self._add_christmas_day(tr("Eerste Kerstdag"))

        # Second Day of Christmas.
        self._add_christmas_day_two(tr("Tweede Kerstdag"))

        # Holi.
        self._add_holi(tr("Holi-Phagwa"))

        if self._year >= 2012:
            # Diwali.
            self._add_diwali(tr("Divali"))

        # Eid al-Fitr.
        self._add_eid_al_fitr_day(tr("Ied-Ul-Fitre"))

        if self._year >= 2012:
            # Eid al-Adha.
            self._add_eid_al_adha_day(tr("Ied-Ul-Adha"))

        if self._year >= 2022:
            # Chinese New Year.
            self._add_chinese_new_years_day(tr("Chinees Nieuwjaar"))


class SR(Suriname):
    pass


class SUR(Suriname):
    pass


class SurinameHinduHolidays(_CustomHinduHolidays):
    # https://web.archive.org/web/20241104221047/https://www.timeanddate.com/holidays/suriname/holi-phagwa
    HOLI_DATES = {
        2015: (MAR, 6),
        2016: (MAR, 23),
        2017: (MAR, 12),
        2018: (MAR, 2),
        2019: (MAR, 21),
        2020: (MAR, 9),
        2021: (MAR, 28),
        2022: (MAR, 18),
        2023: (MAR, 7),
        2024: (MAR, 25),
        2025: (MAR, 14),
        2026: (MAR, 3),
        2027: (MAR, 22),
        2028: (MAR, 11),
        2029: (MAR, 28),
        2030: (MAR, 19),
    }

    # https://web.archive.org/web/20241105043434/https://www.timeanddate.com/holidays/suriname/diwali
    DIWALI_DATES = {
        2014: (OCT, 23),
        2015: (NOV, 11),
        2016: (OCT, 30),
        2017: (OCT, 19),
        2018: (NOV, 7),
        2019: (OCT, 27),
        2020: (NOV, 14),
        2021: (NOV, 4),
        2022: (OCT, 24),
        2023: (NOV, 12),
        2024: (OCT, 31),
        2025: (OCT, 20),
        2026: (NOV, 8),
        2027: (OCT, 28),
        2028: (OCT, 17),
        2029: (NOV, 5),
        2030: (OCT, 25),
    }


class SurinameIslamicHolidays(_CustomIslamicHolidays):
    # https://web.archive.org/web/20241107062349/https://www.timeanddate.com/holidays/suriname/eid-al-fitr
    EID_AL_FITR_DATES_CONFIRMED_YEARS = (2014, 2025)
    EID_AL_FITR_DATES = {
        2014: (JUL, 29),
        2015: (JUL, 18),
        2016: (JUL, 7),
        2017: (JUN, 26),
        2019: (JUN, 5),
        2023: (APR, 22),
        2025: (MAR, 31),
    }

    # https://web.archive.org/web/20241113121535/https://www.timeanddate.com/holidays/suriname/eid-al-adha
    EID_AL_ADHA_DATES_CONFIRMED_YEARS = (2014, 2025)
    EID_AL_ADHA_DATES = {
        2014: (OCT, 5),
        2015: (SEP, 24),
        2016: (SEP, 13),
        2017: (SEP, 2),
        2019: (AUG, 12),
        2023: (JUN, 29),
        2025: (JUN, 7),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from holidays.calendars import _CustomIslamicHolidays
from holidays.calendars.gregorian import JAN, APR, OCT
from holidays.groups import (
    ChristianHolidays,
    IslamicHolidays,
    InternationalHolidays,
    StaticHolidays,
)
from holidays.holiday_base import HolidayBase


class Comoros(
    HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays, StaticHolidays
):
    """Comoros holidays.

    References:
        * <https://en.wikipedia.org/wiki/Public_holidays_in_the_Comoros>
        * [2021 Prophet's Birthday](https://web.archive.org/web/20250712053206/https://scontent-del1-2.xx.fbcdn.net/v/t39.30808-6/471586197_1113861707132176_3696210890184034016_n.jpg?_nc_cat=100&ccb=1-7&_nc_sid=833d8c&_nc_ohc=3hfvuVNGVbMQ7kNvwHvrgbW&_nc_oc=AdnuDzvtcvdBJMoN-qSW7Pnx2fnzj2TFgwMSRklpM_Na-5AaB9y8D0U5BINcxL-vVL0&_nc_zt=23&_nc_ht=scontent-del1-2.xx&_nc_gid=HWC_59SRsyDfwGxku5bqXw&oh=00_AfSk0LCPS7PARPjdX5-vfM_lr7WxZnrkA9jRyrdUFhoyMQ&oe=68736290)
    """

    country = "KM"
    # %s (estimated).
    estimated_label = "%s (estimated)"
    # Independence from France on July 6, 1975.
    start_year = 1976

    def __init__(self, *args, islamic_show_estimated: bool = True, **kwargs):
        """
        Args:
            islamic_show_estimated:
                Whether to add "estimated" label to Islamic holidays name
                if holiday date is estimated.
        """
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(
            self, cls=ComorosIslamicHolidays, show_estimated=islamic_show_estimated
        )
        StaticHolidays.__init__(self, ComorosStaticHolidays)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # New Year's Day.
        self._add_new_years_day("New Year's Day")

        # Cheikh al Maarouf Day.
        self._add_holiday_mar_18("Cheikh al Maarouf Day")

        # Labor Day.
        self._add_labor_day("Labour Day")

        # National Day.
        self._add_holiday_jul_6("National Day")

        if self._year >= 2006:
            # Maore Day.
            self._add_holiday_nov_12("Maore Day")

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Islamic New Year.
        self._add_islamic_new_year_day("Islamic New Year")

        # Prophet's Birthday.
        self._add_mawlid_day("Prophet's Birthday")

        # Isra' and Mi'raj.
        self._add_isra_and_miraj_day("Isra' and Mi'raj")

        # Eid al-Fitr.
        name = "Eid al-Fitr"
        self._add_eid_al_fitr_day(name)
        self._add_eid_al_fitr_day_two(name)
        self._add_eid_al_fitr_day_three(name)

        # Eid al-Adha.
        name = "Eid al-Adha"
        self._add_eid_al_adha_day(name)
        self._add_eid_al_adha_day_two(name)


class KM(Comoros):
    pass


class COM(Comoros):
    pass


class ComorosIslamicHolidays(_CustomIslamicHolidays):
    MAWLID_DATES = {
        2021: (OCT, 19),
    }


class ComorosStaticHolidays:
    """Comoros special holidays.

    References:
        * [Décret N°24-056PR](https://web.archive.org/web/20250717033820/https://munganyo.km/decrees/225)
        * [Décret N°25-008PR](https://web.archive.org/web/20250715223159/https://munganyo.km/decrees/374)
    """

    special_public_holidays = {
        2024: (APR, 13, "National Holiday"),
        2025: (JAN, 30, "Election Partial Day Holiday"),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import SEP, DEC
from holidays.groups import ChristianHolidays, InternationalHolidays, StaticHolidays
from holidays.observed_holiday_base import (
    ObservedHolidayBase,
    TUE_TO_PREV_MON,
    WED_THU_TO_NEXT_FRI,
    SAT_SUN_TO_NEXT_MON,
    SUN_TO_NEXT_MON,
    SUN_TO_NEXT_TUE,
)


class Bahamas(ObservedHolidayBase, ChristianHolidays, InternationalHolidays, StaticHolidays):
    """Bahamas holidays.

    References:
        * <https://en.wikipedia.org/wiki/Public_holidays_in_the_Bahamas>
        * <https://web.archive.org/web/20250421024112/http://laws.bahamas.gov.bs/cms/images/LEGISLATION/PRINCIPAL/1938/1938-0015/PublicHolidaysAct_1.pdf>
        * <https://archive.org/details/majority-rule-public-holiday-act-2013>
        * <https://archive.org/details/randol-fawkes-labour-day-act-2013>
        * <https://web.archive.org/web/20210922210604/http://www.tribune242.com/news/2013/oct/12/national-heroes-day-formally-established/>
        * <https://web.archive.org/web/20190905051207/http://eleutheranews.com/?p=3594>

    Checked With:
        * <https://web.archive.org/web/20250218064125/https://www.bahamashclondon.net/consular-information/public-holidays/>
        * <https://web.archive.org/web/20250427182357/https://bisxbahamas.com/wp-content/uploads/2020/12/Trading-Calendar-2021.pdf>
        * <https://web.archive.org/web/20231210053149/https://publicholidays.la/the-bahamas/2022-dates/>
    """

    country = "BS"
    observed_label = "%s (observed)"
    # Gained Independence on Jul 10, 1973.
    start_year = 1974

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        StaticHolidays.__init__(self, BahamasStaticHolidays)
        kwargs.setdefault("observed_rule", SAT_SUN_TO_NEXT_MON)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # New Year's Day.
        # Pre-2012 Observance:
        # - If TUE, New Year's Day (observed) prev MON.
        # - If WED or THU, New Year's Day (observed) next FRI.
        # 2012 and beyond Observance: If SUN, New Year's Day (observed) next MON (not for SAT).
        self._add_observed(
            self._add_new_years_day("New Year's Day"),
            rule=SUN_TO_NEXT_MON
            if self._year >= 2012
            else SAT_SUN_TO_NEXT_MON + TUE_TO_PREV_MON + WED_THU_TO_NEXT_FRI,
        )

        # Majority Rule Day.
        # Officially made a holiday on Oct 11, 2013 under Majority Rule (Public Holiday) Act 2013.
        if self._year >= 2014:
            self._add_observed(self._add_holiday_jan_10("Majority Rule Day"))

        # Good Friday.
        self._add_good_friday("Good Friday")

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        # Whit Monday.
        self._add_whit_monday("Whit Monday")

        # Randol Fawkes Labour Day.
        # This was simply known as "Labour Day" prior to Randol Fawkes Labour Day Act 2013.
        self._add_holiday_1st_fri_of_jun(
            "Randol Fawkes Labour Day" if self._year >= 2013 else "Labour Day"
        )

        # Independence Day.
        self._add_observed(self._add_holiday_jul_10("Independence Day"))

        # Emancipation Day.
        self._add_holiday_1st_mon_of_aug("Emancipation Day")

        # National Heroes Day.
        # Known as "Discovery Day" prior to 2013, with its date fixed as Oct 12 annually.
        # Got its name changed on Oct 11, 2013 under Majority Rule (Public Holiday) Act 2013.
        # Pre-2013 Observance:
        # - If TUE, Discovery Day (observed) prev MON.
        # - If WED or THU, Discovery Day (observed) next FRI.
        if self._year >= 2013:
            self._add_holiday_2nd_mon_of_oct("National Heroes Day")
        else:
            self._add_observed(
                self._add_columbus_day("Discovery Day"),
                rule=SAT_SUN_TO_NEXT_MON + TUE_TO_PREV_MON + WED_THU_TO_NEXT_FRI,
            )

        # Christmas Holidays Exception Rules.
        # Observance Exception:
        # FRI-SAT -> Boxing Day (observed) next MON.
        # SAT-SUN -> Boxing Day (observed) next MON.
        # SUN-MON -> Christmas Day (observed) next TUE.

        # Christmas Day.
        self._add_observed(self._add_christmas_day("Christmas Day"), rule=SUN_TO_NEXT_TUE)

        # Boxing Day.
        self._add_observed(self._add_christmas_day_two("Boxing Day"))


class BS(Bahamas):
    pass


class BHS(Bahamas):
    pass


class BahamasStaticHolidays:
    special_public_holidays = {
        # https://web.archive.org/web/20250122110613/https://www.bahamas.gov.bs/wps/portal/public/gov/government/notices/national%20holiday%2019th%20september/
        2022: (SEP, 19, "State Funeral of Queen Elizabeth II"),
    }

    special_public_holidays_observed = {
        # New Year's Day observance overflow.
        # This only applies to Pre-2012 observance.
        1979: (DEC, 31, "New Year's Day"),
        1984: (DEC, 31, "New Year's Day"),
        1990: (DEC, 31, "New Year's Day"),
        2001: (DEC, 31, "New Year's Day"),
        2007: (DEC, 31, "New Year's Day"),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, FEB, APR, MAY, JUN, AUG, SEP
from holidays.groups import ChristianHolidays, InternationalHolidays, StaticHolidays
from holidays.observed_holiday_base import (
    ObservedHolidayBase,
    SAT_SUN_TO_NEXT_MON,
    SAT_SUN_TO_NEXT_MON_TUE,
)


class Gibraltar(ObservedHolidayBase, ChristianHolidays, InternationalHolidays, StaticHolidays):
    """Gibraltar holidays.

    References:
        * <https://en.wikipedia.org/wiki/Public_holidays_in_Gibraltar>
        * [2000](https://web.archive.org/web/20250714134510/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/1999s073.pdf)
        * [2001](https://web.archive.org/web/20250714134609/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2000s041.pdf)
        * [2003](https://web.archive.org/web/20250713064759/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2002s094.pdf)
        * [2004](https://web.archive.org/web/20250713065008/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2003s105.pdf)
        * [2005](https://web.archive.org/web/20250713071914/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2004s069.pdf)
        * [2006](https://web.archive.org/web/20250713072021/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2005s111.pdf)
        * [2007](https://web.archive.org/web/20250713072021/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2006s122.pdf)
        * [2008](https://web.archive.org/web/20250713072026/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2007s113.pdf)
        * [2009](https://web.archive.org/web/20250713072221/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2008s073.pdf)
        * [2010](https://web.archive.org/web/20250713072222/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2009s049.pdf)
        * [2011](https://web.archive.org/web/20250713072228/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2010s149.pdf)
        * [2012](https://web.archive.org/web/20250713073334/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2011s189.pdf)
        * [2013](https://web.archive.org/web/20250713162959/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2012s146.pdf)
        * [2014](https://web.archive.org/web/20250713103345/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2013s149.pdf)
        * [2015](https://web.archive.org/web/20250713163521/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2014s203.pdf)
        * [2016](https://web.archive.org/web/20250713164519/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2015s196.pdf)
        * [2017](https://web.archive.org/web/20250713073752/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2016s217.pdf)
        * [2018](https://web.archive.org/web/20250714135250/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2017s224.pdf)
        * [2019](https://web.archive.org/web/20250713071343/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2018s229.pdf)
        * [2020](https://web.archive.org/web/20250714135425/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2019s217.pdf)
        * [2021](https://web.archive.org/web/20250421063152/https://www.gibraltarlaws.gov.gi/legislations/bank-and-public-holidays-order-2020-5593/download)
        * [2022](https://web.archive.org/web/20250414134257/https://www.gibraltarlaws.gov.gi/legislations/bank-and-public-holidays-order-2021-6286/version/21-10-2021/download)
        * [2023](https://web.archive.org/web/20240831014104/https://www.gibraltarlaws.gov.gi/legislations/bank-and-public-holidays-order-2022-6735/version/22-02-2023/download)
        * [2024](https://web.archive.org/web/20250502030403/https://www.gibraltarlaws.gov.gi/legislations/bank-and-public-holidays-order-2023-7215/version/06-11-2023/download)
        * [2025](https://web.archive.org/web/20250714135600/https://www.gibraltarlaws.gov.gi/legislations/bank-and-public-holidays-order-2024-7456/download)
    """

    country = "GI"
    default_language = "en_GB"
    # %s (observed).
    observed_label = tr("%s (observed)")
    # First holiday info available from 2000.
    start_year = 2000
    supported_languages = ("en_GB", "en_US")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        StaticHolidays.__init__(self, GibraltarStaticHolidays)
        kwargs.setdefault("observed_rule", SAT_SUN_TO_NEXT_MON)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # New Year's Day.
        self._add_observed(self._add_new_years_day(tr("New Year's Day")))

        name = (
            # Winter Midterm Bank Holiday.
            tr("Winter Midterm Bank Holiday")
            if self._year >= 2023
            # Commonwealth Day.
            else tr("Commonwealth Day")
        )
        winter_midterm_dts = {
            2024: (FEB, 12),
        }
        if dt := winter_midterm_dts.get(self._year):
            self._add_holiday(name, dt)
        elif self._year >= 2021:
            self._add_holiday_3rd_mon_of_feb(name)
        else:
            self._add_holiday_2nd_mon_of_mar(name)

        # Good Friday.
        self._add_good_friday(tr("Good Friday"))

        # Easter Monday.
        self._add_easter_monday(tr("Easter Monday"))

        if self._year >= 2013:
            # Workers' Memorial Day.
            name = tr("Workers' Memorial Day")
            workers_memorial_dts = {
                2013: (APR, 26),
                2015: (APR, 27),
            }
            if dt := workers_memorial_dts.get(self._year):
                self._add_holiday(name, dt)
            else:
                self._add_observed(self._add_holiday_apr_28(name))

        # May Day.
        name = tr("May Day")
        may_day_dts = {
            2007: (MAY, 7),
            2008: (MAY, 5),
            2009: (MAY, 4),
        }
        if dt := may_day_dts.get(self._year):
            self._add_holiday(name, dt)
        else:
            self._add_observed(self._add_holiday_may_1(name))

        # Spring Bank Holiday.
        name = tr("Spring Bank Holiday")
        spring_bank_dts = {
            2012: (JUN, 4),
            2022: (JUN, 2),
        }
        if dt := spring_bank_dts.get(self._year):
            self._add_holiday(name, dt)
        else:
            self._add_holiday_last_mon_of_may(name)

        name = (
            # King's Birthday.
            tr("King's Birthday")
            if self._year >= 2023
            # Queen's Birthday.
            else tr("Queen's Birthday")
        )
        sovereign_birthday_dts = {
            2000: (JUN, 19),
            2001: (JUN, 18),
            2006: (JUN, 19),
            2007: (JUN, 18),
            2012: (JUN, 18),
            2013: (JUN, 17),
            2017: (JUN, 19),
            2019: (JUN, 17),
            2023: (JUN, 19),
            2024: (JUN, 17),
        }
        if dt := sovereign_birthday_dts.get(self._year):
            self._add_holiday(name, dt)
        else:
            self._add_holiday_2_days_past_2nd_sat_of_jun(name)

        self._add_holiday_last_mon_of_aug(
            # Summer Bank Holiday.
            tr("Summer Bank Holiday")
            if 2002 <= self._year <= 2007
            # Late Summer Bank Holiday.
            else tr("Late Summer Bank Holiday")
        )

        # Gibraltar National Day.
        name = tr("Gibraltar National Day")
        national_day_dts = {
            2016: (SEP, 5),
            2017: (SEP, 4),
        }
        if dt := national_day_dts.get(self._year):
            self._add_holiday(name, dt)
        else:
            self._add_observed(self._add_holiday_sep_10(name))

        self._add_observed(
            # Christmas Day.
            self._add_christmas_day(tr("Christmas Day")),
            rule=SAT_SUN_TO_NEXT_MON_TUE,
        )

        self._add_observed(
            # Boxing Day.
            self._add_christmas_day_two(tr("Boxing Day")),
            rule=SAT_SUN_TO_NEXT_MON_TUE,
        )


class GI(Gibraltar):
    pass


class GIB(Gibraltar):
    pass


class GibraltarStaticHolidays(StaticHolidays):
    """Gibraltar special holidays.

    References:
        * [Evacuation Commemoration Day, 2015](https://web.archive.org/web/20250714174537/https://www.gibraltarlaws.gov.gi/uploads/legislations/banking-and-financial-dealings/B&P%20Holidays/2015s112.pdf)
    """

    special_public_holidays = {
        # Tercentenary Holiday.
        2004: (AUG, 4, tr("Tercentenary Holiday")),
        # Bank Holiday.
        2009: (JAN, 12, tr("Bank Holiday")),
        # Queen's Diamond Jubilee.
        2012: (JUN, 5, tr("Queen's Diamond Jubilee")),
        # Evacuation Commemoration Day.
        2015: (SEP, 7, tr("Evacuation Commemoration Day")),
        # 75th Anniversary of VE Day.
        2020: (MAY, 8, tr("75th Anniversary of VE Day")),
        # Queen's Platinum Jubilee.
        2022: (JUN, 3, tr("Platinum Jubilee")),
        # Special King's Coronation Bank Holiday.
        2023: (MAY, 8, tr("Special King's Coronation Bank Holiday")),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from datetime import date

from holidays.helpers import _normalize_tuple


class StaticHolidays:
    """Helper class for special and substituted holidays support.

    Populates special and substituted holidays related data from
    an external class.
    """

    def __init__(self, cls) -> None:
        for attribute_name in cls.__dict__.keys():
            # Special holidays.
            if attribute_name.startswith("special_") and (
                value := getattr(cls, attribute_name, None)
            ):
                setattr(self, attribute_name, value)
                self.has_special_holidays = True

            # "Substituted" labels.
            elif attribute_name.startswith("substituted_") and (
                value := getattr(cls, attribute_name, None)
            ):
                setattr(self, attribute_name, value)

        # Populate substituted holidays from adjacent years.
        self.weekend_workdays = set()
        for special_public_holidays in getattr(self, "special_public_holidays", {}).values():
            for special_public_holiday in _normalize_tuple(special_public_holidays):
                # Normally, special holiday is a 3 item tuple: (month, day, name).
                if len(special_public_holiday) < 4:  # Skip non-substituted holidays.
                    continue

                # Handle cross-year substituted holidays.
                if len(special_public_holiday) == 5:  # The fifth element is the year.
                    _, _, from_month, from_day, from_year = special_public_holiday
                    self.weekend_workdays.add(date(from_year, from_month, from_day))

                self.has_substituted_holidays = True



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see CONTRIBUTORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars import (
    _CustomBuddhistHolidays,
    _CustomChineseHolidays,
    _CustomIslamicHolidays,
    _CustomHinduHolidays,
)
from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.groups import (
    BuddhistCalendarHolidays,
    ChineseCalendarHolidays,
    ChristianHolidays,
    HinduCalendarHolidays,
    InternationalHolidays,
    IslamicHolidays,
    StaticHolidays,
)
from holidays.observed_holiday_base import ObservedHolidayBase, SUN_TO_NEXT_WORKDAY


class Singapore(
    ObservedHolidayBase,
    BuddhistCalendarHolidays,
    ChineseCalendarHolidays,
    ChristianHolidays,
    HinduCalendarHolidays,
    InternationalHolidays,
    IslamicHolidays,
    StaticHolidays,
):
    """Singapore holidays.

    References:
        * [Wikipedia](https://en.wikipedia.org/wiki/Public_holidays_in_Singapore)
        * [Holidays Act 1998](https://web.archive.org/web/20250405061431/https://sso.agc.gov.sg/Act/HA1998)
        * [Ministry of Manpower](https://web.archive.org/web/20250616105633/https://mom.gov.sg/employment-practices/public-holidays)

    Limitations:
        * Prior to 1969: holidays are estimated.
        * Prior to 2000: holidays may not be accurate.
        * 2024 and later: the following four moving date holidays (whose exact
            date is announced yearly) are estimated, and so denoted:
            * Hari Raya Puasa
            * Hari Raya Haji
            * Vesak Day
            * Deepavali
    """

    country = "SG"
    # %s (estimated).
    estimated_label = tr("%s (estimated)")
    default_language = "en_SG"
    # %s (observed, estimated).
    observed_estimated_label = tr("%s (observed, estimated)")
    # %s (observed).
    observed_label = tr("%s (observed)")
    supported_languages = ("en_SG", "en_US", "th")

    def __init__(self, *args, islamic_show_estimated: bool = True, **kwargs):
        """
        Args:
            islamic_show_estimated:
                Whether to add "estimated" label to Islamic holidays name
                if holiday date is estimated.
        """
        BuddhistCalendarHolidays.__init__(self, cls=SingaporeBuddhistHolidays, show_estimated=True)
        ChineseCalendarHolidays.__init__(self, cls=SingaporeChineseHolidays, show_estimated=True)
        ChristianHolidays.__init__(self)
        HinduCalendarHolidays.__init__(self, cls=SingaporeHinduHolidays)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(
            self, cls=SingaporeIslamicHolidays, show_estimated=islamic_show_estimated
        )
        StaticHolidays.__init__(self, cls=SingaporeStaticHolidays)
        # Implement Section 4(2) of the Holidays Act:
        # "if any day specified in the Schedule falls on a Sunday,
        # the day next following not being itself a public holiday
        # is declared a public holiday in Singapore."
        kwargs.setdefault("observed_rule", SUN_TO_NEXT_WORKDAY)
        kwargs.setdefault("observed_since", 1998)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self) -> None:
        dts_observed = set()

        # New Year's Day.
        dts_observed.add(self._add_new_years_day(tr("New Year's Day")))

        # Chinese New Year.
        name = tr("Chinese New Year")
        dts_observed.add(self._add_chinese_new_years_day(name))  # type: ignore[arg-type]
        dts_observed.add(self._add_chinese_new_years_day_two(name))  # type: ignore[arg-type]

        # Eid al-Fitr.
        dts_observed.update(self._add_eid_al_fitr_day(tr("Hari Raya Puasa")))
        if self._year <= 1968:
            # Second Day of Eid al-Fitr.
            self._add_eid_al_fitr_day_two(tr("Second Day of Hari Raya Puasa"))

        # Eid al-Adha.
        dts_observed.update(self._add_eid_al_adha_day(tr("Hari Raya Haji")))

        # Good Friday.
        self._add_good_friday(tr("Good Friday"))

        if self._year <= 1968:
            # Holy Saturday.
            self._add_holy_saturday(tr("Holy Saturday"))

            # Easter Monday.
            self._add_easter_monday(tr("Easter Monday"))

        # Labor Day.
        dts_observed.add(self._add_labor_day(tr("Labour Day")))

        # Vesak Day.
        dts_observed.add(self._add_vesak(tr("Vesak Day")))  # type: ignore[arg-type]

        # National Day.
        dts_observed.add(self._add_holiday_aug_9(tr("National Day")))

        # Deepavali.
        dts_observed.add(self._add_diwali(tr("Deepavali")))  # type: ignore[arg-type]

        # Christmas Day.
        dts_observed.add(self._add_christmas_day(tr("Christmas Day")))

        if self._year <= 1968:
            # Boxing Day.
            self._add_christmas_day_two(tr("Boxing Day"))

        if self.observed:
            self._populate_observed(dts_observed)


class SG(Singapore):
    pass


class SGP(Singapore):
    pass


class SingaporeBuddhistHolidays(_CustomBuddhistHolidays):
    VESAK_DATES = {
        2001: (MAY, 7),
        2002: (MAY, 26),
        2003: (MAY, 15),
        2004: (JUN, 2),
        2005: (MAY, 22),
        2006: (MAY, 12),
        2007: (MAY, 31),
        2008: (MAY, 19),
        2009: (MAY, 9),
        2010: (MAY, 28),
        2011: (MAY, 17),
        2012: (MAY, 5),
        2013: (MAY, 24),
        2014: (MAY, 13),
        2015: (JUN, 1),
        2016: (MAY, 21),
        2017: (MAY, 10),
        2018: (MAY, 29),
        2019: (MAY, 19),
        2020: (MAY, 7),
        2021: (MAY, 26),
        2022: (MAY, 15),
        2023: (JUN, 2),
        2024: (MAY, 22),
        2025: (MAY, 12),
        2026: (MAY, 31),
    }


class SingaporeChineseHolidays(_CustomChineseHolidays):
    LUNAR_NEW_YEAR_DATES_CONFIRMED_YEARS = (2001, 2026)
    LUNAR_NEW_YEAR_DATES = {
        2006: (JAN, 30),
        2007: (FEB, 19),
    }


class SingaporeHinduHolidays(_CustomHinduHolidays):
    # Deepavali
    DIWALI_DATES = {
        2001: (NOV, 14),
        2002: (NOV, 3),
        2003: (OCT, 23),
        2004: (NOV, 11),
        2005: (NOV, 1),
        2006: (OCT, 21),
        2007: (NOV, 8),
        2008: (OCT, 27),
        2009: (NOV, 15),
        2010: (NOV, 5),
        2011: (OCT, 26),
        2012: (NOV, 13),
        2013: (NOV, 2),
        2014: (OCT, 22),
        2015: (NOV, 10),
        2016: (OCT, 29),
        2017: (OCT, 18),
        2018: (NOV, 6),
        2019: (OCT, 27),
        2020: (NOV, 14),
        2021: (NOV, 4),
        2022: (OCT, 24),
        2023: (NOV, 12),
        2024: (OCT, 31),
        2025: (OCT, 20),
        2026: (NOV, 8),
    }


class SingaporeIslamicHolidays(_CustomIslamicHolidays):
    EID_AL_ADHA_DATES_CONFIRMED_YEARS = (2001, 2026)
    EID_AL_ADHA_DATES = {
        2001: (MAR, 6),
        2002: (FEB, 23),
        2003: (FEB, 12),
        2010: (NOV, 17),
        2014: (OCT, 5),
        2015: (SEP, 24),
        2016: (SEP, 12),
        2018: (AUG, 22),
        2022: (JUL, 10),
        2023: (JUN, 29),
        2024: (JUN, 17),
        2025: (JUN, 7),
    }

    EID_AL_FITR_DATES_CONFIRMED_YEARS = (2001, 2026)
    EID_AL_FITR_DATES = {
        2002: (DEC, 6),
        2006: (OCT, 24),
        2019: (JUN, 5),
        2022: (MAY, 3),
        2023: (APR, 22),
        2025: (MAR, 31),
        2026: (MAR, 21),
    }


class SingaporeStaticHolidays:
    """Singapore special holidays.

    References:
        * <https://web.archive.org/web/20241015024728/https://www.mom.gov.sg/newsroom/press-releases/2015/sg50-public-holiday-on-7-august-2015>
        * <https://web.archive.org/web/20240809195048/https://www.mom.gov.sg/newsroom/press-releases/2020/0624-public-holiday-on-polling-day---10-july-2020>
        * <https://web.archive.org/web/20241113193000/https://www.mom.gov.sg/newsroom/press-releases/2023/0822-public-holiday-on-polling-day---1-sep-2023>
        * <https://web.archive.org/web/20250424145037/https://www.mom.gov.sg/newsroom/press-releases/2025/0415-public-holiday-on-polling-day_3-may-2025>
    """

    # Polling Day.
    polling_day_name = tr("Polling Day")

    special_public_holidays = {
        2001: (NOV, 3, polling_day_name),
        2006: (MAY, 6, polling_day_name),
        2011: (MAY, 7, polling_day_name),
        2015: (
            # SG50 Public Holiday.
            (AUG, 7, tr("SG50 Public Holiday")),
            (SEP, 11, polling_day_name),
        ),
        2020: (JUL, 10, polling_day_name),
        2023: (SEP, 1, polling_day_name),
        2025: (MAY, 3, polling_day_name),
    }

    special_public_holidays_observed = {
        # Eid al-Adha.
        2007: (JAN, 2, tr("Hari Raya Haji")),
    }
